# -*- coding: utf-8 -*-
"""Hybrid time series classifiers."""
__all__ = [
    "HIVECOTEV1",
    "HIVECOTEV2",
]

from sktime.classification.hybrid._hivecote_v1 import HIVECOTEV1
from sktime.classification.hybrid._hivecote_v2 import HIVECOTEV2
