# -*- coding: utf-8 -*-
"""
A place for misc/incomplete developer code.

This is not an official sktime package. The contents stored here are not guaranteed to
be properly reviewed, and can change at any time with no deprecation.
"""
