# -*- coding: utf-8 -*-
"""Distance based time series regressors."""

__all__ = ["KNeighborsTimeSeriesRegressor"]

from sktime.regression.distance_based._time_series_neighbors import (
    KNeighborsTimeSeriesRegressor,
)
