#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
"""Implement interval based time series regression estimators."""

__author__ = ["Markus Löning"]
__all__ = ["TimeSeriesForestRegressor"]

from sktime.regression.interval_based._tsf import TimeSeriesForestRegressor
