# -*- coding: utf-8 -*-
"""Module exports: Series type checkers, converters and mtype inference."""

from sktime.datatypes._table._check import check_dict as check_dict_Table
from sktime.datatypes._table._convert import convert_dict as convert_dict_Table
from sktime.datatypes._table._examples import example_dict as example_dict_Table
from sktime.datatypes._table._examples import (
    example_dict_lossy as example_dict_lossy_Table,
)
from sktime.datatypes._table._examples import (
    example_dict_metadata as example_dict_metadata_Table,
)
from sktime.datatypes._table._registry import MTYPE_LIST_TABLE, MTYPE_REGISTER_TABLE

__all__ = [
    "check_dict_Table",
    "convert_dict_Table",
    "MTYPE_LIST_TABLE",
    "MTYPE_REGISTER_TABLE",
    "example_dict_Table",
    "example_dict_lossy_Table",
    "example_dict_metadata_Table",
]
